// Filename		: Serializable.h
// Author		: Siddharth Barman
// Date			: 06 Aug 2005
// Description	: Declares interface ISerializable. This interface must be imple
//				  mented by classes which want to be serializable. These classes
//				  must register with  a class implementing the ISerializer inter
//				  face. 
//				  Also declares IObjectFactory interface which is required to be
//				  implemented by the same class which implements ISerializable o
//				  r by a different class. The Serializer will make use of 
//				  IObjectFactory to create new instances of a class during de-se
//				  rialization.
//------------------------------------------------------------------------------
#pragma once

#include <afx.h>
#include <afxcoll.h>
//------------------------------------------------------------------------------

enum PropertyType
{
	Blank,
	Simple,
	SimpleList,
	Complex,
	ComplexList
};
//------------------------------------------------------------------------------
class CProperty;
class IObjectFactory;
//------------------------------------------------------------------------------

class ISerializable
{
public:
	virtual			~ISerializable(){};
	
	virtual int		GetProperties(CStringList& properties) = 0;

	virtual bool	GetPropertyValue(const CString& sProperty, 
									 CProperty& sValue) = 0;

	virtual bool	SetPropertyValue(const CString& sProperty, 
									 CProperty& sValue) = 0;

	virtual bool	HasMultipleInstances() = 0;									 
	virtual CString GetClassName() = 0;
	virtual CString GetID() = 0;
};
//------------------------------------------------------------------------------

// This class is a wrapper for a property which belongs to a class. It is capabl
// e of representing simple string property, an object implementing ISerializabl
// e, a list of string values and a list of ISerializable objects. Each property
// is identified by a property name (string).
class CProperty
{
public:
	// utility function for everyone
	static void CopyStringList(CStringList& dst, const CStringList& src)
	{
		// Make a copy of the passed in string list
		POSITION pos = src.GetHeadPosition();
		CString sValue;
		
		dst.RemoveAll();

		while(pos)
		{			
			sValue = src.GetNext(pos);
			dst.AddTail(sValue);
		}		
	}

	// utility function for everyone
	static void CopyPtrList(CPtrList& dst, const CPtrList& src)
	{
		// Make a copy of the passed in string list
		POSITION pos = src.GetHeadPosition();
		
		dst.RemoveAll();

		while(pos)
		{				
			dst.AddTail((void*)src.GetNext(pos));
		}		
	}	
	
	CProperty(IObjectFactory* factory = NULL)
	{		
		m_object	= NULL;
		m_type		= Blank;
		m_factory	= factory;
	}

	CProperty(const CString& sName)
	{
		m_sName = sName;
	}

	CProperty(const CString& sName, const CString& sValue)
	{
		m_sName		= sName;
		m_sValue	= sValue;
		m_object	= NULL;
		m_type		= Simple;
	}

	CProperty(const CString& sName, ISerializable* object)
	{
		m_sName		= sName;
		m_object	= object;
		m_type		= Complex;
	}

	IObjectFactory* GetFactory() const
	{
		return m_factory;
	}

	void SetFactory(IObjectFactory* factory)
	{
		m_factory = factory;
	}

	CString GetName() const
	{
		return m_sName;
	}

	void SetName(const CString& sValue)
	{
		m_sName = sValue;
	}

	CString GetStringValue() const
	{
		return m_sValue;
	}

	ISerializable* GetObject() const
	{
		return m_object;
	}

	CStringList& GetStringList()
	{
		return m_values;
	}

	CPtrList& GetObjectList()
	{
		return m_objects;
	}

	CProperty& operator=(const CString& rhs)
	{
		m_sValue = rhs;
		m_object = NULL;
		m_type	 = Simple;
		return *this;
	}

	CProperty& operator=(const CStringList& rhs)
	{
		// Make a copy of the passed in string list
		CopyStringList(m_values, rhs);			
		m_object = NULL;
		m_type	 = SimpleList;
		return *this;
	}

	CProperty& operator=(const CPtrList& rhs)
	{
		// Make a copy of the passed in string list
		CopyPtrList(m_objects, rhs);
		m_object = NULL;
		m_type	 = ComplexList;
		return *this;
	}

	CProperty& operator=(ISerializable* rhs)
	{
		m_sValue = _T("");
		m_object = rhs;
		m_type	 = Complex;
		return *this;
	}
	
	operator CString() const
	{
		return m_sValue;
	}

	operator ISerializable*() const
	{
		return m_object;
	}

	PropertyType GetType() const
	{
		return m_type;
	}
	
	void SetType(PropertyType type)
	{
		m_type = type;
	}

	CString GetTypeString() const
	{
		CString sType;	

		switch(m_type)
		{
		case Blank:
			sType = _T("Blank");
			break;

		case Simple:
			sType = _T("Simple");
			break;

		case SimpleList:
			sType = _T("SimpleList");
			break;

		case Complex:
			sType = _T("Complex");
			break;

		case ComplexList:
			sType = _T("ComplexList");
			break;

		default:
			sType = "Blank";		
		}

		return sType;
	}

	static PropertyType Parse(const CString& sType)
	{
		PropertyType type = Blank;

		if(sType == _T("Simple"))
			type = Simple;
		else if(sType == _T("SimpleList"))
			type = SimpleList;
		else if(sType == _T("Complex"))
			type = Complex;
		else if(sType == _T("ComplexList"))
			type = ComplexList;
		
		return type;
	}	

private:
	PropertyType	m_type;
	CString			m_sName;
	CString			m_sValue;	// single simple (string )value
	ISerializable*	m_object;	// string complex value
	CStringList		m_values;	// list of simple (string) values
	CPtrList		m_objects;	// list of complex values
	IObjectFactory* m_factory;
};
//------------------------------------------------------------------------------

class IObjectFactory
{
public:
	virtual ISerializable* Create() = 0; // Factory function
	virtual void		   Destroy(ISerializable*) = 0;

};
//------------------------------------------------------------------------------