// Filename		: XMLSerializer.cpp
// Author		: Siddharth Barman
// Date			: Aug 06 2005
// Description	: Implementation file for 

//------------------------------------------------------------------------------
#include "StdAfx.h"
#include "XmlSerializer.h"

//------------------------------------------------------------------------------

CXMLSerializer::CXMLSerializer()
{
	m_doc.CreateInstance(CLSID_DOMDocument26);
}
//------------------------------------------------------------------------------

CXMLSerializer::CXMLSerializer(const CString& sFile, const CString& sAppName, 
							   bool bLoad)
{
	try	
	{
		m_doc.CreateInstance(CLSID_DOMDocument26);
		SetFile(sFile);
		SetApplicationName(sAppName);

		if(bLoad)
		{
			// read in the xml file
			ReadFile();
		}
		else
		{
			// create a new xml structure
			MSXML::IXMLDOMNodePtr rootnode = CreateChildNode(m_sAppName);
			m_doc->appendChild(rootnode);
		}
	}
	catch(_com_error err)
	{
		throw CXMLException(err.ErrorMessage());
	}
}
//------------------------------------------------------------------------------

 CXMLSerializer::~CXMLSerializer()
{
	if(!m_sFile.IsEmpty())
	{
		_bstr_t bsFile(m_sFile);
		m_doc->save(bsFile);
	}

	m_doc = NULL;
}
//------------------------------------------------------------------------------

void CXMLSerializer::ReadFile()
{	
	ASSERT(m_sFile.IsEmpty() == FALSE);

	_bstr_t bsFile(m_sFile);
	m_doc->load(bsFile);	
}
//------------------------------------------------------------------------------

void CXMLSerializer::SetFile(const CString& sValue)
{
	m_sFile = sValue;
}
//------------------------------------------------------------------------------

CString CXMLSerializer::GetFile()
{
	return m_sFile;
}
//------------------------------------------------------------------------------

void CXMLSerializer::SetApplicationName(const CString& sValue)
{
	m_sAppName = sValue;
}
//------------------------------------------------------------------------------

CString CXMLSerializer::GetApplicationName()
{
	return m_sAppName;
}
//------------------------------------------------------------------------------

// Returns the plural form of a word e.g. cat becomes cats
CString CXMLSerializer::GetPlural(const CString& sWord)
{
	CString sResult;
	TCHAR ch = sWord.GetAt(sWord.GetLength()-1);
	
	if(ch == _T('s') || ch == _T('S'))
	{
		// e.g. address become address[e]s
		sResult.Format(_T("%ses"), sWord);
	}
	else if(ch == _T('y') || ch == _T('Y'))
	{
		// e.g. city becomes cities
		sResult.Format(_T("%sies"), sWord);
	}
	else
	{
		// e.g. dog becomes dogs
		sResult.Format(_T("%ss"), sWord);
	}
	return sResult;
}

//------------------------------------------------------------------------------
bool CXMLSerializer::SerializeObject(ISerializable* obj, 
									 MSXML::IXMLDOMNode* parent)
{
	// To serialize the object correctly, we need to get its identity and then 
	// get the property list after which we query for value of each property.
	// We also need to check if the XML file has a node for this object, this 
	// will be done based on the ID of the object <object id="xxx">. 

	ASSERT(obj != NULL);
	if(obj == NULL)
		return false;

	CString sClass	= obj->GetClassName();			// e.g. <contact>	
	CString sId		= obj->GetID();					// e.g. <contact id="1">	
	bool	bMulti	= obj->HasMultipleInstances();	// if multiple instances has
													// to be stored
	
	// lets check if the object has already been persisted
	_bstr_t bsXPath;
	CString sPath;
	CString sOuterNodeName;
	MSXML::IXMLDOMNodePtr node;
	MSXML::IXMLDOMNodePtr refnode = parent;

	if(bMulti)
	{
		// do some english lang processing for getting the plural form
		sOuterNodeName = GetPlural(sClass);
		
		// e.g. <contacts>
		//			<contact>		
		if(parent == NULL)
		{
			// reference is the root			
			sPath.Format(_T("//%s/%s/%s[id='%s']"), m_sAppName, sOuterNodeName,
						 sClass, sId);			
		}
		else
		{
			// reference is the parent node
			sPath.Format(_T("%ss/%s[id='%s']"), sOuterNodeName, sClass, sId);	
		}
	}
	else
	{
		// e.g. <contact> 
		if(parent == NULL)
		{
			// reference is the root
			sPath.Format(_T("//%s/%s[id='%s']"), m_sAppName, sClass, sId);
		}
		else
		{
			// reference is the parent node
			sPath.Format(_T("%s[id='%s']"), sClass, sId);
		}
	}

	bsXPath = sPath;

	MSXML::IXMLDOMNodePtr rootnode = m_doc->GetfirstChild();
	MSXML::IXMLDOMNodePtr outernode; // e.g. <contacts>

	if(refnode == NULL)
	{
		// since no reference node is present, the reference node is the 
		// root node.
		refnode = rootnode;
	}
	
	if(NULL == refnode)
	{
		// this is probably a brand new document without any contents
		refnode = CreateChildNode(m_sAppName, rootnode);
		m_doc->appendChild(refnode);
	}

	node = refnode->selectSingleNode(bsXPath);

	if(node == NULL)
	{
		// it is possible that the parent node e.g. <contacts> itself is
		// not there. We need to create it.
		if(bMulti)
		{
			CString sOuterNodeName;

			sOuterNodeName = sClass + _T("s"); // e.g. contact+s = contacts
			sPath.Format(_T("//%s/%s"), m_sAppName, sOuterNodeName);
			bsXPath = sPath;			
			
			outernode = refnode->selectSingleNode(bsXPath);

			if(outernode == NULL)
			{
				// this node does not exist, we need to create it				
				outernode = CreateChildNode(sOuterNodeName, rootnode);
				refnode->appendChild(outernode);				
			}			
		}
	}
	
	CStringList props;
	CProperty	property;
	CString		sProperty;
	CString		sValue;
	int			nProps;
	int			nIndex;
	
	nProps = obj->GetProperties(props); // property count

	for(nIndex = 0; nIndex < nProps; nIndex++)
	{
		sProperty	= props.GetAt( props.FindIndex(nIndex) );
		property.SetName(sProperty);

		if(obj->GetPropertyValue(sProperty, property))
		{
			// create the node if NULL
			if(node == NULL)
			{
				// this is the entity node e.g. <contact>
				if(bMulti == false)
				{					
					node = CreateChildNode(sClass, refnode);
				}
				else
				{
					node = CreateChildNode(sClass, outernode);
				}

				// set the ID attribute
				if(!sId.IsEmpty())
				{
					SetAttributeValue(node, _T("ID"), sId);
				}
			}

			switch(property.GetType())
			{
			case Simple:
				sValue = property.GetStringValue();				
				SetProperty(node, property); // persist sValue in the xml
				break;

			case SimpleList:
				{
					// first we create the outer node e.g. <phonenumbers>
					CString sOuterNodeName = GetPlural(sProperty);
					CStringList* values = &property.GetStringList();
					MSXML::IXMLDOMNodePtr outernode;
			
					// attach the outernode
					outernode = CreateChildNode(sOuterNodeName, node);
					node->appendChild(outernode);
					
					// iterate through all the list of values and store them
					for(int nIndex = 0; nIndex < values->GetCount(); nIndex++)
					{
						sValue = values->GetAt(values->FindIndex(nIndex));
						CProperty singleprop(sProperty, sValue);
						SetProperty(outernode, singleprop, true); // true means
						// we want a new node to be created		
					}				
				}
				break;
			
			case Complex:				
				SerializeObject(property.GetObject(), node);
				break;
			
			case ComplexList:
				{
					// first we create the outer node e.g. <phonenumbers>
					CString sOuterNodeName = GetPlural(sProperty);
					CPtrList* values = &property.GetObjectList();
					MSXML::IXMLDOMNodePtr outernode;
			
					// attach the outernode
					outernode = CreateChildNode(sOuterNodeName, node);
					node->appendChild(outernode);
					
					// iterate through all the list of values and store them
					for(int nIndex = 0; nIndex < values->GetCount(); nIndex++)
					{
						ISerializable* pObject = (ISerializable*)
												 (values->GetAt(values->FindIndex(nIndex)));
						CProperty singleprop(sProperty, pObject);
						SerializeObject(pObject, outernode);							
					}				
				}
				break;
			}			
		}
	}

	_bstr_t bsPath(m_sFile);
	m_doc->save(bsPath);
	return true;
}
//------------------------------------------------------------------------------

// Serializes a single object.
bool CXMLSerializer::Serialize(ISerializable* obj)
{
	return SerializeObject(obj, NULL);
}
//------------------------------------------------------------------------------

// Serialized a list of objects.
bool CXMLSerializer::Serialize(const CPtrList& objects)
{
	bool	 bResult = true;
	POSITION pos = objects.GetHeadPosition();

	while(pos && bResult)
	{
		ISerializable* object = (ISerializable*)objects.GetNext(pos);
		bResult = SerializeObject(object);
	}
	
	return bResult;
}
//------------------------------------------------------------------------------

// Returns number of items deserialized. In case of fatal error, returns 
// DESERIALIZATION_FAIL. Created objects are added to the objList
int CXMLSerializer::DeserializeObject(IObjectFactory* factory, CPtrList& objList, 
									  MSXML::IXMLDOMNode* parent)
{
	ASSERT(factory != NULL);
	if(NULL == factory)
	{
		return DESERIALIZATION_FAIL;
	}

	// create an empty instance of the object to de-serializable
	ISerializable* pEmptyObject = factory->Create(); // this needs to to be
													 // destroyed

	ASSERT(pEmptyObject != NULL);
	if(NULL == pEmptyObject)
	{		
		factory->Destroy(pEmptyObject);
		return DESERIALIZATION_FAIL;
	}

	CString		sClass = pEmptyObject->GetClassName();
	CString		sPath;
	CString		sId;
	CString		sPropName;
	CString		sPropValue;
	CString		sPropType;
	CStringList sProps;
	int			nProperties;
	int			nObjects = 0;
		
	nProperties = pEmptyObject->GetProperties(sProps);

	if(nProperties == 0)
	{
		TRACE(_T("\nZero properties for Class:"));
		TRACE(sClass);
		factory->Destroy(pEmptyObject);
		return 0;
	}

	bool		bMulti = pEmptyObject->HasMultipleInstances();
	_bstr_t		bsXPath;

	if(bMulti)
	{
		// e.g. <contacts>
		//			<contact>
		if(parent == NULL)
		{
			// reference is root
			sPath.Format(_T("//%s/%ss/%s"), m_sAppName, sClass, sClass);
		}
		else
		{
			// reference is parent node
			sPath.Format(_T("%ss/%s"), sClass, sClass);
		}
	}
	else
	{
		// e.g. <contact> 
		if(parent == NULL)
		{
			// reference is root
			sPath.Format(_T("//%s/%s/%s"), m_sAppName, sClass, sId);
		}
		else
		{
			// reference is parent node
			sPath.Format(_T("%s"), sClass);			
		}
	}

	bsXPath = sPath;
	MSXML::IXMLDOMNodePtr	  refnode;
	MSXML::IXMLDOMNodeListPtr nodes;

	if(NULL == parent)
	{
		refnode = m_doc->GetfirstChild();
	}
	else
	{
		refnode = parent;
	}

	if(NULL == refnode)
	{
		return 0; // even the parent node does not exist
				  // This is not an error, all it means is that
				  // no such objects have been deserialized
	}

	nodes =  refnode->selectNodes(bsXPath);
	
	if(nodes == NULL)
	{
		TRACE(_T("\nNo nodes for XPATH: "));
		TRACE(sPath);
		factory->Destroy(pEmptyObject);
		return 0;
	}

	int nLength = nodes->Getlength();
	
	// loop for iterating over the class nodes e.g. <contact>
	for(int nIndex = 0; nIndex < nodes->Getlength(); nIndex++)
	{
		MSXML::IXMLDOMNodePtr node = nodes->Getitem(nIndex);

		if(node != NULL)
		{
			ISerializable* pObj = factory->Create();
		
			GetAttributeValue(node, _T("ID"), sId); // get the ID

			// loop through all the properties
			for(int nPropIndex = 0; nPropIndex < nProperties; nPropIndex++)
			{
				sPropName = sProps.GetAt(sProps.FindIndex(nPropIndex));
			
				CProperty property(sPropName);

				if(pObj->GetPropertyValue(sPropName, property))
				{
					switch(property.GetType())
					{
					case Simple:						
						if(GetProperty(node, property))
						{					
							pObj->SetPropertyValue(sPropName, property);
						}
						break;

					case SimpleList:
						if(GetProperty(node, property))
						{
							pObj->SetPropertyValue(sPropName, property);
						}

						break;

					case Complex:
						{
						// get a list of new objects for the complex object at hand
						CPtrList list;
						DeserializeObject(property.GetFactory(), list, node);

						// note: newly created object are present in the list. These object
						// will be destroyed by the outer class which will accept these 
						// objects as a property-value. In case the outer class wants to
						// hold on to these objects, they can do so and destroy them later.
						// The outer object class might also decide to immediately destroy
						// the passed in object. This is true only for Complex types.
						
						// iterate through the complex objects and add them to the previous
						// (parent) object
						for(int n=0; n < list.GetCount(); n++)
						{
							CProperty newproperty;
							newproperty.SetFactory(property.GetFactory());
							newproperty = (ISerializable*)list.GetAt(list.FindIndex(n));
							
							// update the parent
							// the parent is now responsible for the object lifetime
							pObj->SetPropertyValue(sPropName, newproperty); 																	
						}					
						}
						break;

					case ComplexList:
						{
						// first locate the outernode, then walk over the subnodes
						// get a list of new objects for each complex object at hand
						CString sOuterNode = GetPlural(sPropName);
						MSXML::IXMLDOMNodeListPtr subnodes;
						MSXML::IXMLDOMNodePtr complexnode;
						bsXPath = sOuterNode;
						
						subnodes = node->selectNodes(bsXPath);

						if(NULL == subnodes)
						{
							break;
						}
						
						for(int n = 0; n < subnodes->Getlength(); n++)
						{
							complexnode = subnodes->Getitem(n);
							CPtrList list;												
							DeserializeObject(property.GetFactory(), list, complexnode);

							CProperty newproperty(sPropName);
							newproperty.SetFactory(property.GetFactory());
							newproperty = list;
							pObj->SetPropertyValue(sPropName, newproperty);
						}												
						} // case ComplexList
						break;
					} // switch					
				}				
			}
			
			objList.AddTail(pObj);
			nObjects++;
		}
	}
	factory->Destroy(pEmptyObject);
	return nObjects;
}
//------------------------------------------------------------------------------

// Returns number of items deserialized. In case of fatal error, returns 
// DESERIALIZATION_FAIL. Created objects are added to the objList
int CXMLSerializer::Deserialize(IObjectFactory* factory, CPtrList& objList)
{
	return DeserializeObject(factory, objList, NULL);
}
//------------------------------------------------------------------------------

// If a parentnode is supplied, it will create a new node and add it as a child
// node of the parent node. If parentnode is NULL, it will create a new node and
// return it. Imp: accept the returned node and Release it.
MSXML::IXMLDOMNode* CXMLSerializer::CreateChildNode(const CString& sNodeName, 
													MSXML::IXMLDOMNode* parent)
{	
	MSXML::IXMLDOMNodePtr node;
	
	_variant_t	vtType(MSXML::NODE_ELEMENT);
	_bstr_t		bsNode(sNodeName);
	_bstr_t		bsNamespace;

	node = m_doc->createNode(vtType, bsNode, bsNamespace);
		
	ASSERT(node != NULL);
	if(node == NULL)
	{
		return NULL;
	}

	if(parent != NULL)
	{
		parent->appendChild(node);
	}
	
	node->AddRef();
	
	return node;
}
//------------------------------------------------------------------------------

// Pass a node pointer, property name & property value. e.g. <student> and 
// property is fname, will read text value from <fname> under student. 
// Also works for string lists.
bool CXMLSerializer::GetProperty(MSXML::IXMLDOMNode* node, CProperty& property)								 								 
{
	ASSERT(node != NULL);
	if(node == NULL)
	{
		return false;
	}

	CString						sNodeName;
	MSXML::IXMLDOMNodePtr		propnode;	// property node
	MSXML::IXMLDOMNodeListPtr	subnodes;	// only for lists
	CStringList*				pList;		// only for lists			
	_bstr_t						bsXPath;	
	_bstr_t						bsValue;
	CString						sPropType;
	CString						sValue;
	PropertyType				type;

	ISerializable* pComplexProperty = NULL;
	type = property.GetType();

	switch(type)
	{
	case Simple:
		bsXPath = property.GetName();
		propnode = node->selectSingleNode(bsXPath);

		if(propnode == NULL)
		{		
			TRACE(_T("\nProperty node was not found."));
			return false;
		}
		
		//if(!GetAttributeValue(propnode, _T("Type"), sPropType))
		//	return false;

		bsValue = propnode->Gettext();	
		property	= (LPCTSTR)bsValue;
		
		return true;

	case SimpleList:
		sNodeName = GetPlural(property.GetName());
		bsXPath = sNodeName;
		subnodes = node->selectNodes(bsXPath);
		
		pList = &(property.GetStringList());

		if(subnodes != NULL)
		{
			MSXML::IXMLDOMNodePtr singlenode;
							
			for(int n = 0; n < subnodes->Getlength(); n++)
			{
				singlenode = subnodes->Getitem(n);
				bsValue =  singlenode->Gettext();
				pList->AddTail((LPCTSTR)bsValue);
			}
		}
		return true;
		break;
	
	default:
		return false;
	}	
}
//------------------------------------------------------------------------------	

// Pass a node pointer, property name & property value. Will create the property
// node if required. e.g. <student> and property is fname, will create <fname>
// under student if needed and set the text value to the passed in value.
bool CXMLSerializer::SetProperty(MSXML::IXMLDOMNode* node, 
								const CProperty& property,
								bool bCreateNewNode)
{
	ASSERT(node != NULL);
	if(NULL == node)
	{
		return false;
	}

	CString sProp = property.GetName();
	MSXML::IXMLDOMNodePtr		propnode;	// property node
	CString						sValue = property;
	_bstr_t						bsXPath(sProp);	
	_bstr_t						bsValue(sValue);
	
	if(false == bCreateNewNode)
	{
		// node existence checking is done only if the function was
		// told NOT to enforce new node creation
		propnode = node->selectSingleNode(bsXPath);
	}

	if(NULL == propnode)
	{
		// this node does not exists, we need to create it
		_variant_t	vtType(MSXML::NODE_ELEMENT);
		_bstr_t		bsNode(sProp);
		_bstr_t		bsNamespace;

		propnode = m_doc->createNode(vtType, bsNode, bsNamespace);
		
		ASSERT(propnode != NULL);
		if(propnode == NULL)
		{
			return false;
		}
		
		node->appendChild(propnode); // attach the node

		// this is not really required
		//SetAttributeValue(propnode, _T("Type"), property.GetTypeString());
	}
	
	
	propnode->Puttext(bsValue);

	return true;
}
//------------------------------------------------------------------------------

bool CXMLSerializer::SetAttributeValue(MSXML::IXMLDOMNode* node, 
								  const CString& sAttName, 
								  const CString& sValue)
{
	ASSERT(node != NULL);
	if(node == NULL)
	{
		return false;
	}

	try
	{
        MSXML::IXMLDOMNodePtr attnode;
						
		_bstr_t		bsNode(sAttName);
		_bstr_t		bsValue(sValue);

		attnode = m_doc->createAttribute(bsNode);
			
		ASSERT(attnode != NULL);
		if(attnode == NULL)
		{
			return false;
		}
		
		attnode->Puttext(bsValue);
		node->Getattributes()->setNamedItem(attnode);
		
		return true;	
	}
	catch(_com_error& err)
	{
		TRACE(err.ErrorMessage());
		return false;
	}
	catch(...)
	{
		TRACE(UNSPEC_ERROR);
		return false;
	}
}
//------------------------------------------------------------------------------

bool CXMLSerializer::GetAttributeValue(MSXML::IXMLDOMNode* node, 
									   const CString& sAttName, CString& sValue)
{
	ASSERT(node != NULL);
	if(node == NULL)
	{
		return false;
	}

	try
	{
        MSXML::IXMLDOMNodePtr attnode;
		MSXML::IXMLDOMNamedNodeMapPtr attributes;
						
		_bstr_t		bsName(sAttName);
		_bstr_t		bsValue(sValue);

		attributes = node->Getattributes();

		if(attributes == NULL)
		{
			return false;
		}

		attnode = attributes->getNamedItem(bsName);
		if(attnode == NULL)
		{
			TRACE(_T("Attribute "));
			TRACE(sAttName);
			TRACE(_T(" not found."));
			return false;
		}
		
		bsValue = attnode->Gettext();
		sValue	= (LPCTSTR)bsValue;
		
		return true;	
	}
	catch(_com_error& err)
	{
		TRACE(err.ErrorMessage());
		return false;
	}
	catch(...)
	{
		TRACE(UNSPEC_ERROR);
		return false;
	}
}

//  ******** CXMLException Class *******
CXMLException::CXMLException(const CString& sMsg)
{
	m_sMessage = sMsg;
}
//------------------------------------------------------------------------------

CXMLException::CXMLException(const CXMLException& lhs)
{
	m_sMessage = lhs.m_sMessage;
}
//------------------------------------------------------------------------------

CString CXMLException::ToString()
{
	return m_sMessage;
}
//------------------------------------------------------------------------------

BOOL CXMLException::GetErrorMessage(LPTSTR lpszError, UINT nMaxError, 
									PUINT pnHelpContext)
{
	if((int)nMaxError > m_sMessage.GetLength())
	{
		lstrcpy(lpszError, m_sMessage);
		return TRUE;
	}
	else
	{
		return FALSE;
	}	
}
//------------------------------------------------------------------------------

int CXMLException::ReportError(UINT nType, UINT nMessageID)
{
	CString sMsg;

	if(nMessageID != 0)
	{
		return AfxMessageBox(nMessageID, nType, 0);
	}
	else
	{
		return AfxMessageBox(m_sMessage, nType, 0);
	}	
}
//------------------------------------------------------------------------------